﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Reversi
{
    // AI 1号（探索：ミニマックス法、評価：単純に石の数）
    class AI1 : AI
    {
        public AI1(int depth, int order) : base(depth, order){}

        // 次の手を取得する
        public override Move getNextMove(Board board)
        {
            // 次の手を考える
            readMinMax(board, 0);

            return bestMove;
        }

        // ミニマックス法による先読み (再帰)
        public int readMinMax(Board board, int depth)
        {
            // 先読み深さの末端に到達したら評価値を返す
            if (depth == maxDepth)
            {
                // 評価関数は石の数
                int eval = board.countStone(myOrder);

                return eval;
            }

            // 次の局面を列挙
            List<Move> nextMoves = board.enumNextMoves();

            // 自分の手番なら最も自分に有利な手を選択（自分にとっての最善手）
            // 相手の手番なら最も自分に不利な手を選択（相手にとっての最善手）
            int best = (board.turnHolder == myOrder) ? int.MinValue : int.MaxValue;
            for (int i = 0; i < nextMoves.Count; i++)
            {
                // 再帰呼び出し
                Board nextBoard = board.copy();
                nextBoard.doMove(nextMoves[i]);
                int eval = readMinMax(nextBoard, depth + 1);

                if ((board.turnHolder == myOrder) && (eval > best))
                {
                    best = eval;
                    if (depth == 0)
                    {
                        bestMove = nextMoves[i];
                    }
                }
                if ((board.turnHolder != myOrder) && (eval < best))
                {
                    best = eval;
                    if (depth == 0)
                    {
                        bestMove = nextMoves[i];
                    }
                }
            }
            return best;
        }
    }
}
