﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Reversi
{
    // AI 2号（探索：ミニマックス法、評価：石の位置による重みづけ）
    class AI2 : AI
    {
        public AI2(int depth, int order) : base(depth, order){}

        // 評価の重みづけ
        private int[,] WEIGHT = {
            {+50, -10, +5, +5, +5, +5, -10, +50},
            {-10, -10, -2, -2, -2, -2, -10, -10},
            { +5,  -2, +2, +1, +1, +2,  -2,  +5},
            { +5,  -2, +1, +1, +1, +1,  -2,  +5},
            { +5,  -2, +1, +1, +1, +1,  -2,  +5},
            { +5,  -2, +2, +1, +1, +2,  -2,  +5},
            {-10, -10, -2, -2, -2, -2, -10, -10},
            {+50, -10, +5, +5, +5, +5, -10, +50}
        };

        // 評価関数
        private int evalFunc(Board board)
        {
            int eval = 0;

            for (int x = 0; x < Board.SIZE; x++)
            {
                for (int y = 0; y < Board.SIZE; y++)
                {
                    if (board.stone[x, y] == myOrder)
                    {
                        eval += WEIGHT[x, y];
                    }
                    else if (board.stone[x, y] == -myOrder)
                    {
                        eval -= WEIGHT[x, y];
                    }
                }
            }
            return eval;
        }

        // 次の手を取得する
        public override Move getNextMove(Board board)
        {
            // 次の手を考える
            readMinMax(board, 0);

            return bestMove;
        }

        // ミニマックス法による先読み (再帰)
        public int readMinMax(Board board, int depth)
        {
            // 先読み深さの末端に到達したら評価値を返す
            if (depth == maxDepth)
            {
                // 評価関数
                int eval = evalFunc(board);

                return eval;
            }

            // 次の局面を列挙
            List<Move> nextMoves = board.enumNextMoves();

            // 自分の手番なら最も自分に有利な手を選択（自分にとっての最善手）
            // 相手の手番なら最も自分に不利な手を選択（相手にとっての最善手）
            int best = (board.turnHolder == myOrder) ? int.MinValue : int.MaxValue;
            for (int i = 0; i < nextMoves.Count; i++)
            {
                // 再帰呼び出し
                Board nextBoard = board.copy();
                nextBoard.doMove(nextMoves[i]);
                int eval = readMinMax(nextBoard, depth + 1);

                if ((board.turnHolder == myOrder) && (eval > best))
                {
                    best = eval;
                    if (depth == 0)
                    {
                        bestMove = nextMoves[i];
                    }
                }
                if ((board.turnHolder != myOrder) && (eval < best))
                {
                    best = eval;
                    if (depth == 0)
                    {
                        bestMove = nextMoves[i];
                    }
                }
            }
            return best;
        }
    }
}
