﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Reversi
{
    public partial class Form1 : Form
    {
        // デバッグ用
        const bool isDebug = false;//true; // trueならCPU対CPU
        AI cpu2AI;                 // CPU2の人工知能

        // 升目の幅
        const int BOX_WIDTH = 60;
        // 石の直径
        const int STONE_SIZE = 50;
        
        // 盤面データ
        Board board = new Board();
        // 着手とひっくり返しデータ
        ReverseFlag reverse = new ReverseFlag();
        // 対局は開始したか？
        bool isStarted = false;

        // 人間の先手/後手
        int you = Board.FIRST_MOVE;
        // CPUの先手/後手
        int cpu = Board.SECOND_MOVE;

        // CPUのAI
        AI cpuAI;
        
        // 初期化
        public Form1()
        {
            InitializeComponent();

            comboMove.SelectedIndex = 0;
            textDepth.Text = "6";

            board.reset();
            reverse.reset();
            drawBoard();
            textTurn.Text = "対局前";
            textTurn.ForeColor = Color.Green; 
        }

        // 盤面の描画
        private void drawBoard()
        {
            Bitmap canvas = new Bitmap(pictureBoard.Width, pictureBoard.Height);
            Graphics g = Graphics.FromImage(canvas);

            // 緑色の盤面
            Rectangle rect = new Rectangle(0, 0, pictureBoard.Width-1, pictureBoard.Height-1);
            g.FillRectangle(Brushes.Green, rect);

            // 升目の黒線
            int px1 = 0;
            int px2 = BOX_WIDTH * Board.SIZE;
            int py1 = 0;
            int py2 = BOX_WIDTH * Board.SIZE;
            for (int x = 0; x <= Board.SIZE; x++)
            {
                int px = x * BOX_WIDTH;
                g.DrawLine(Pens.Black, px, py1, px, py2);
            }
            for (int y = 0; y <= Board.SIZE; y++)
            {
                int py = y * BOX_WIDTH;
                g.DrawLine(Pens.Black, px1, py, px2, py);
            }

            // 石
            for (int x = 0; x < Board.SIZE; x++)
            {
                for (int y = 0; y < Board.SIZE; y++)
                {
                    // 今回着手した石の升目は赤で
                    if (reverse.stone[x, y] == ReverseFlag.MOVE_STONE)
                    {
                        int px = x * BOX_WIDTH + 1;
                        int py = y * BOX_WIDTH + 1;
                        int pw = BOX_WIDTH -2;
                        g.FillRectangle(Brushes.Red, px, py, pw, pw);
                    }
                    // 今回ひっくり返った石の升目を黄色で
                    else if (reverse.stone[x, y] == ReverseFlag.REVERSED_STONE)
                    {
                        int px = x * BOX_WIDTH + 1;
                        int py = y * BOX_WIDTH + 1;
                        int pw = BOX_WIDTH - 2;
                        g.FillRectangle(Brushes.Yellow, px, py, pw, pw);
                    }
                    // 先手の石（黒）
                    if (board.stone[x, y] == Board.FIRST_MOVE)
                    {
                        int px = x * BOX_WIDTH + (BOX_WIDTH - STONE_SIZE) / 2;
                        int py = y * BOX_WIDTH + (BOX_WIDTH - STONE_SIZE) / 2;
                        g.FillEllipse(Brushes.Black, px, py, STONE_SIZE, STONE_SIZE);
                    }
                    // 後手の石（白）
                    else if (board.stone[x, y] == Board.SECOND_MOVE)
                    {
                        int px = x * BOX_WIDTH + (BOX_WIDTH - STONE_SIZE) / 2;
                        int py = y * BOX_WIDTH + (BOX_WIDTH - STONE_SIZE) / 2;
                        g.FillEllipse(Brushes.White, px, py, STONE_SIZE, STONE_SIZE);
                    }
                }
            }

            g.Dispose();
            pictureBoard.Image = canvas;

            // 石の個数を表示
            textYou.Text = board.countStone(you).ToString();
            textComputer.Text = board.countStone(cpu).ToString();

            // どちらが手番持ちかを表示
            textTurn.Text = (board.turnHolder == you) ? "あなた" : "CPU";
            textTurn.ForeColor = (board.turnHolder == you) ? Color.Red : Color.Blue;
        }

        // 対局開始/停止ボタン
        private void buttonStart_Click(object sender, EventArgs e)
        {
            // 停止
            if (isStarted)
            {
                isStarted = false;
                buttonStart.Text = "開始";
                comboMove.Enabled = true;
                textDepth.Enabled = true;
                textTurn.Text = "停止";
                textTurn.ForeColor = Color.Green; 
            }
            // 開始
            else
            {
                // 先読み深さの設定をチェック
                int depth;
                try
                {
                    depth = int.Parse(textDepth.Text);
                    if (depth < 1) throw new Exception();
                }
                catch
                {
                    MessageBox.Show("読みの深さの設定が無効です!");
                    return;
                }
                // 先手/後手のチェック
                you = ( comboMove.SelectedIndex == 0) ? Board.FIRST_MOVE : Board.SECOND_MOVE;
                cpu = (comboMove.SelectedIndex == 0) ? Board.SECOND_MOVE : Board.FIRST_MOVE;
                
                // CPUのAIを生成
                cpuAI = new AI2(depth, cpu);
                if (isDebug)
                {
                    cpuAI = new AI1(6, cpu);
                    cpu2AI = new AI2(6, you);
                }

                // 盤面のリセット
                board.reset();
                reverse.reset();
                drawBoard();
                
                // CPUが先手の場合
                if (cpu == Board.FIRST_MOVE)
                {
                    // CPUの手番タスク
                    Task.Run(() =>
                    {
                        taskCpuTurn();
                    });
                }
                else if (isDebug)
                {
                    // CPU2の手番タスク
                    Task.Run(() =>
                    {
                        taskCpu2Turn();
                    });
                }

                isStarted = true;
                buttonStart.Text = "停止";
                comboMove.Enabled = false;
                textDepth.Enabled = false;
            }
        }

        // 盤面のクリック（着手の操作）
        private void pictureBoard_MouseClick(object sender, MouseEventArgs e)
        {
            // 対局中かチェック
            if (!isStarted)
            {
                return;
            }
            // 人間の手番かチェック
            if (board.turnHolder == cpu)
            {
                MessageBox.Show("あなたの手番ではありません!");
                return;
            }

            // 着手位置の計算
            int px = e.X;
            int py = e.Y;
            int x = px / BOX_WIDTH;
            int y = py / BOX_WIDTH;
            Move move = new Move(x, y);

            // 有効な着手かチェック
            if (board.isAvailableMove(move))
            {
                // 着手する
                reverse = board.doMove(move);
                // 盤面の描画
                drawBoard();
                // 終了判定
                if (checkGameFinished()) return;

                // CPUの手番タスク
                Task.Run(() =>
                {
                    taskCpuTurn();　
                });
            }
            else
            {
                MessageBox.Show("無効な着手です!");
            }
        }
        
        // CPUの手番タスク関数
        void taskCpuTurn()
        {
            // 次の手を考える
            Move next = cpuAI.getNextMove(board);
            // 着手
            reverse = board.doMove(next);

            this.Invoke((Action)(() =>
            {
                // 盤面の描画
                drawBoard();
                // 終了判定
                if (checkGameFinished())
                {
                    return;
                }
                // あなたが着手可能かの判定
                if(!board.hasAvailableMove())
                {
                    // あなたはパス
                    if (!isDebug)
                    {
                        MessageBox.Show("着手できません。パスします");
                    }
                    board.doMove(Reversi.Move.PASS);

                    // CPUの手番タスク
                    Task.Run(() =>
                    {
                        taskCpuTurn();
                    });
                }
                else if (isDebug)
                {
                    // あなたの代わりにCPU2の手番タスク
                    Task.Run(() =>
                    {
                        taskCpu2Turn();
                    });
                }
            }));　
        }

        // CPU2の手番タスク関数 (デバッグ用)
        void taskCpu2Turn()
        {
            // 次の手を考える
            Move next = cpu2AI.getNextMove(board);
            // 着手
            reverse = board.doMove(next);

            this.Invoke((Action)(() =>
            {
                // 盤面の描画
                drawBoard();
                // 終了判定
                if (checkGameFinished())
                {
                    return;
                }
                // CPU1が着手可能かの判定
                if (!board.hasAvailableMove())
                {
                    // CPU1はパス
                    board.doMove(Reversi.Move.PASS);

                    // CPU2の手番タスク
                    Task.Run(() =>
                    {
                        taskCpu2Turn();
                    });
                }
                else
                {
                    // CPU1の手番タスク
                    Task.Run(() =>
                    {
                        taskCpuTurn();
                    });
                }
            }));
        }

        // 終了判定
        bool checkGameFinished()
        {
            if (board.hasFinished())
            {
                // 勝ち負け判定とメッセージ表示
                if (board.winner == you)
                {
                    MessageBox.Show("あなたの勝ちです");
                }
                else if (board.winner == cpu)
                {
                    MessageBox.Show("あなたの負けです");
                }
                else
                {
                    MessageBox.Show("引き分けです");
                }
                // 停止ボタンを押す
                buttonStart_Click(new object(), new EventArgs());
                textTurn.Text = "終了";
                textTurn.ForeColor = Color.Green; 

                return true;
            }
            else
            {
                return false;
            }
        }
    }
}
